/* Copyright (C) 2002, by Rupert Paget (rpaget@vision.ee.ethz.ch),
 * all rights reserved.
 *
 * This software is copyrighted as noted below. It may be freely copied,
 * modified, and redistributed, provided that the copyright notice is
 * preserved on all copies.
 *
 * There is no warranty or other guarantee of fitness for this software,
 * it is provided solely "as is". Bug reports or fixes may be sent
 * to the author, who may or may not act on them as he desires.
 *
 * You may not include this software in a commercial program or other
 * software product without the expressed permission of the copyright owner.
 *
 * If you modify this software, you should include a notice giving the
 * name of the person performing the modification, the date of modification,
 * and the reason for such modification.
 */

#include "Glib.h"

/**************************************************************************/

FILE *gfopen(char const *file_name, char const *mode)
{
  FILE *fp;
  
  if ((fp = fopen(file_name, mode)) == NULL) {
    cerr << "Cannot open " << file_name << " - bye!" << endl;
    exit(1);
  }
  return fp;
}

/**************************************************************************/

char *gstrcpy(char *s1, char const *s2)
{
  register char *p = s1;

  while((*p++ = *s2++));
  while(*--p != '.');
  *p = '\0';
  return s1;
}

/**************************************************************************/

char *gstradd(char *s1, char const *s2)
{
  register char *p = s1;
  
  while(*++p);
  while((*p++ = *s2++));
  return s1;
}

/**************************************************************************/

char *gstrnew(char const *s1, char const *s2)
{
  char *p=(char *)malloc(STRING_LENGTH);
  
  gstrcpy(p,s1);
  gstradd(p,s2);
  gstradd(p,".");
  gstradd(p,gstrgetend(s1));
  return p;
}

/**************************************************************************/

const char *gstrgetend(char const *s1)
{
  register const char *p = s1;
  
  while(*p++);
  while(*--p != '.');
  return ++p;
}

/**************************************************************************/

char gstrcmpend(char *s1, char const *s2)
{
  register char *p = s1;
  
  while(*p++);
  while(*--p != '.');
  while(*++p){
    if ((*p < *s2) && (*p + 32 != *s2)) break;
    if ((*p > *s2) && (*p != *s2 + 32)) break;
    s2++;
  }
  return *p == '\0';
}

/**************************************************************************/

int gstrnum(char const *s)
{
  char *p=(char *)malloc(STRING_LENGTH);
  register char *pp=p;
  
  while(*s++);
  while(*--s != '_');
  while((*pp++ = *++s));
  while(*--pp != '.');
  *pp = '\0';
  
  return atoi(p);
}

/**************************************************************************/

void *gcalloc(int n, unsigned sizeof_something)
{
  void *p;
  
  if ((p = calloc(n, sizeof_something)) == NULL) {
    cerr << endl << "ERROR: calloc() failed - bye!" << endl << endl;
    exit(1);
  }
  return p;
}

/**************************************************************************/

void gpause()
{
  char c;
  
  cout << "Please press RETURN to continue ";
  cin >> c;
  cout << endl;
}

/**************************************************************************/

void grunerror(char const error_text[])
  /* Numerical Recipes standard error handler */
{
  cerr << endl << "run-time error..." << endl << error_text << endl;
  cerr << "...now exiting to system..." << endl << endl;
  exit(1);
}

/**************************************************************************/

char *gvector(int r, int size)
  /* allocate a vector with subscript range v[0..r-1] */
{
  char *v;
  
  v = new char [r*size];
  if (!v) grunerror("allocation failure in gvector()");
  return v;
}

/**************************************************************************/

char **gmatrix(int r, int c, int size)
  /* allocate a matrix with subscript range m[0..r-1][0..c-1] */
{
  char **m;
  
  /* allocate pointers to rows */
  m = new char* [r*sizeof(char *)];
  if (!m) grunerror("allocation failure 1 in matrix()");
  
  /* allocate rows and set pointers to them */
  m[0] = new char [r*c*size];
  if (!m[0]) grunerror("allocation failure 2 in matrix()");
  
  for(int i=1; i<r; i++) m[i]= m[i-1]+c*size;
  
  /* return pointer to array of pointers to rows */
  return m;
}

/**************************************************************************/

char ***gtensor(int r, int c, int d, int size)
  /* allocate a 3tensor with range t[0..r-1][0..c-1][0..d-1] */
{
  char ***t;
  
  /* allocate pointers to pointers to rows */
  t = new char** [r*sizeof(char**)];
  if (!t) grunerror("allocation failure 1 in gtensor()");
  
  /* allocate pointers to rows and set pointers to them */
  t[0] = new char* [r*c*sizeof(char*)];
  if (!t[0]) grunerror("allocation failure 2 in gtensor()");
  
  /* allocate rows and set pointers to them */
  t[0][0] = new char [r*c*d*size];
  if (!t[0][0]) grunerror("allocation failure 3 in gtensor()");
  
  for(int j=1; j<c; j++) t[0][j] = t[0][j-1]+d*size;
  for(int i=1; i<r; i++) {
    t[i] = t[i-1]+c;
    t[i][0] = t[i-1][0]+c*d*size;
    for(int j=1; j<c; j++) t[i][j]=t[i][j-1]+d*size;
  }
  
  /* return pointer to array of pointers to rows */
  return t;
}

/**************************************************************************/

void gvectorDelete(char *v)
  /* delete vector */
{
  delete [] v;
  v = 0;
}

/**************************************************************************/

void gmatrixDelete(char **m)
  /* delete matrix */
{
  delete [] *m;
  delete [] m;
  m = 0;
}

/**************************************************************************/

void gtensorDelete(char ***t)
  /* delete tensor */
{
  delete [] **t;
  delete [] *t;
  delete [] t;
  t = 0;
}

/**************************************************************************/

