/* Copyright (C) 2002, by Rupert Paget (rpaget@vision.ee.ethz.ch),
 * all rights reserved.
 *
 * This software is copyrighted as noted below. It may be freely copied,
 * modified, and redistributed, provided that the copyright notice is
 * preserved on all copies.
 *
 * There is no warranty or other guarantee of fitness for this software,
 * it is provided solely "as is". Bug reports or fixes may be sent
 * to the author, who may or may not act on them as he desires.
 *
 * You may not include this software in a commercial program or other
 * software product without the expressed permission of the copyright owner.
 *
 * If you modify this software, you should include a notice giving the
 * name of the person performing the modification, the date of modification,
 * and the reason for such modification.
 */

#include <iostream>
#include <sys/types.h>
#include <time.h>
#include <list>
#include <vector>
#include "ColourImage.h"
#include "GreyImage.h"
#include "Option.h"
#include "Neighbourhood.h"
extern "C" {
#include <nr.h>
}

using namespace std;

#define pMax 255
// #define dMax 195075    // 3 x pMax^2, for Euclidean distance
#define dMax 765       // 3 x pMax, for Manhattan distance
#define ERROR 0.0      // This can be varied to any positive value

BasicImage::byte pUpdateFn(int cO, int rO, GreyImage &imgProb, int increment);
vector<int> *SimilarPixels(ColourImage &img, GreyImage &imgMask,
			   bool mask, float err);
void IndexArray(list<int> &index, int dx, int dy, int cols, int rows,
		GreyImage &imgMask, bool mask, vector<int> &imgIndex);
bool getImgPos(int &c, int &r, list<int> &index, int cols, int rows,
	       GreyImage &imgMask, bool mask);


int main(int argc, const char *argv[]) {
  
  /* Warm up random number generator */
  long iseed;
  time(&iseed);
  iseed = -iseed;
  for (int i = 0; i < 1000; ++i) ran3(&iseed);

  /* Read in arguments */
  Option opt;
  opt.parseArgs(argc, argv);
  while(opt.InputFile()) {
    
    /* Set neighbourhood */
    Neighbourhood n(opt.getNeighOrder(),opt.getSquare());
    
    /* Input training texture */
    ColourImage imgIn(opt.getInputFile(),opt.getLevels());
    int rowsIn = imgIn.getHeight();
    int colsIn = imgIn.getWidth();

    /* Input mask image */
    bool maskIn = opt.getMaskIn();
    GreyImage imgInMask(0,0,opt.getLevels());
    if (maskIn) {
      imgInMask.load(opt.getInputFileMask());
      int rowsInMask = imgInMask.getHeight();
      int colsInMask = imgInMask.getWidth();
      if ((rowsIn != rowsInMask) || (colsIn != colsInMask)) {
	cerr << "Size of mask image does NOT match input texture image";
	cerr << endl;
	exit (1);
      }
    }
    
    /* Set up images */
    int rowsOut = opt.getRows();
    int colsOut = opt.getCols();
    if (rowsOut == 0) rowsOut = rowsIn;
    if (colsOut == 0) colsOut = colsIn;
    opt.setRows(rowsOut);
    opt.setCols(colsOut);
    ColourImage imgOut(colsOut,rowsOut,opt.getLevels()); //Output texture
    GreyImage imgProb(colsOut,rowsOut);                 //Probability map
    imgProb.clear();
    
    /* Set scale */
    int t = 0;
    int scale = 1;
    int treeMax = opt.getTreeMax();
    for (; t <= treeMax; ++t) {
      scale *= 2;
      n.scaleUp();
      if ((scale > rowsIn) && (scale > colsIn)) break;
      if ((scale > rowsOut) && (scale > colsOut)) break;
    }
    if (t < treeMax) treeMax = t;
    opt.setTreeMax(treeMax);

    /* Set output file */
    if (opt.OutputFile()) continue;
    
    /* Malloc variables */
    int nLength = n.getlength();
    int *rAn = new int [nLength]; //Neighbourhood adjustment for cyclic
    int *cAn = new int [nLength]; //Neighbourhood adjustment for cyclic
    bool *pOkN = new bool [nLength];
    float *pForN = new float [nLength];
    BasicImage::byte **imgOutN = new BasicImage::byte* [nLength];
    imgOutN[0] = new BasicImage::byte [nLength*3];
    for (int i = 1; i < nLength; ++i)  imgOutN[i] = imgOutN[i-1]+3;

    /* Input image indexing */
    vector<int> *imgInIndex = SimilarPixels(imgIn,imgInMask,maskIn,ERROR);

    /* Output image indexing */
    int *imgOutIndex = new int [colsOut*rowsOut];
    
    /* Start iteration from scale treeMax down */
    vector<int> rInMin;
    vector<int> cInMin;
    for (t = treeMax; t >= 0; --t){
      scale /= 2;
      n.scaleDown();    
      
      /* If output images are toroidal then in order for the texture to */
      /* wrap properly, we have to shrink the overlapping neighbourhood */
      /* to meet the defined sampling points */
      int rOnDelta = 0;
      int cOnDelta = 0;
      if (t && opt.getCyclic()) {
	if (rowsOut % scale) rOnDelta = scale - (rowsOut % scale);
	if (colsOut % scale) cOnDelta = scale - (colsOut % scale);
      }
      
      /* Determine total number of points to iterate */
      int rO, cO;
      float probTotal = 0.0;
      for (rO = 0; rO < rowsOut; rO += scale)
      for (cO = 0; cO < colsOut; cO += scale)
	probTotal += pMax-imgProb.p(cO,rO);
      
      /* Random loop through output image */
      int pIncrement = pMax/opt.getIterations()+1;
      while (probTotal > 0) {
	float p = ran3(&iseed)*probTotal;
	for (rO = 0; (rO < rowsOut) && (p > 0.0); rO += scale)
	for (cO = 0; (cO < colsOut) && (p > 0.0); cO += scale)
	  p -= pMax-imgProb.p(cO,rO);
	if (rO > 0) rO -= scale;
	if (cO > 0) cO -= scale;
	
	/* Update probabilities */
	BasicImage::byte pUpdate = pUpdateFn(cO,rO,imgProb,pIncrement);
	imgProb.setp(cO,rO,pUpdate+imgProb.p(cO,rO));
	probTotal -= pUpdate;
	
	/* Setup constant Neighbourhood vectors */
	int nPos = -1;
	int dx, dy;
	list<int> index;
	while (n.neigh(dx,dy)) {
	  ++nPos;
	  int rOn = rO+dy;
	  int cOn = cO+dx;
	  rAn[nPos] = 0;
	  cAn[nPos] = 0;
	  
	  /* If output images are toroidal */
	  if (opt.getCyclic()) {
	    if (rOn < 0) {
	      rOn = (rOn + rOnDelta) % rowsOut;
	      if (rOn < 0) rOn += rowsOut;
	      rAn[nPos] = rOnDelta;
	    } else if (rOn >= rowsOut) {
	      rOn = (rOn - rOnDelta) % rowsOut;
	      rAn[nPos] = -rOnDelta;
	    }
	    if (cOn < 0) {
	      cOn = (cOn + cOnDelta) % colsOut;
	      if (cOn < 0) cOn += colsOut;
	      cAn[nPos] = cOnDelta;
	    } else if (cOn >= colsOut) {
	      cOn = (cOn - cOnDelta) % colsOut;
	      cAn[nPos] = -cOnDelta;
	    }
	  }
	  /* Else check suitability */
	  else if ((rOn < 0) || (rOn >= rowsOut) ||
		   (cOn < 0) || (cOn >= colsOut)) {
	    pOkN[nPos] = false;
	    continue;
	  }
	  
	  /* Get neighbour pixel probability */
	  BasicImage::byte pn = imgProb.p(cOn,rOn);
	  if (pn == 0) {
	    pOkN[nPos] = false;
	    continue;
	  }
	  
	  /* Probability factors */
	  pOkN[nPos] = true;
	  pForN[nPos] = float(pn)/pMax;
	  
	  /* Store neighbourhood variables */
	  for (int band = 0; band < 3; ++band)
	    imgOutN[nPos][band] = imgOut.getBand(band)->p(cOn,rOn);
	  
	  /* Update input image indexes */
	  IndexArray(index,dx+cAn[nPos],dy+rAn[nPos],colsIn,rowsIn,imgInMask,
		     maskIn,imgInIndex[imgOutIndex[cOn+rOn*colsOut]]);
	}	    
	
	/* Search input texture for pixel with minumum */
	/* neighbourhood distance */
	int rI, cI;
	float NeighMin = FLT_MAX;
	while (getImgPos(cI,rI,index,colsIn,rowsIn,imgInMask,maskIn)) {
	  nPos = -1;
	  float Neigh = 0.0;
	  while (n.neigh(dx,dy,Neigh,NeighMin)) {
	    ++nPos;
	    int rIn = rI+dy+rAn[nPos];
	    int cIn = cI+dx+cAn[nPos];

	    /* Probability factor */
	    float pFor = pForN[nPos];
	    // pFor *= pFor; // For Euclidean distance
	    
	    /* Check suitability */
	    if (!pOkN[nPos]) continue;
	    if ((rIn < 0) || (rIn >= rowsIn) ||
		(cIn < 0) || (cIn >= colsIn) ||
		(maskIn && !imgInMask.p(cIn,rIn))) {
	      Neigh += dMax*pFor;
	      continue;
	    }
	    
	    /* Calculate Manhatten neighbourhood distance */
	    /* Calculate Euclidean neighbourhood distance */
	    int d;
	    for (int band = 0; band < 3; ++band) {
	      d = imgOutN[nPos][band]-imgIn.getBand(band)->p(cIn,rIn);
	      if (d < 0) Neigh -= d*pFor;
	      else Neigh += d*pFor;
	      // Neigh += d*d*pFor;
	    }
	  }
	  
	  /* Store neighbour pixel distance */
	  if (Neigh < NeighMin) {
	    NeighMin = Neigh;
	    rInMin.clear();
	    cInMin.clear();
	    rInMin.push_back(rI);
	    cInMin.push_back(cI);
	  } else if (Neigh == NeighMin) {
	    rInMin.push_back(rI);
	    cInMin.push_back(cI);
	  }
	}
	
	/* Sample new output pixel */
	int vLength = rInMin.size();
	if (vLength == 1) {
	  rI = rInMin[0];
	  cI = cInMin[0];
	} else {
	  int i = int(ran3(&iseed)*vLength);
	  if (i == vLength) --i;
	  rI = rInMin[i];
	  cI = cInMin[i];
	}
	
	/* Set new output pixel */
	for (int band = 0; band < 3; ++band)
	  imgOut.getBand(band)->setp(cO,rO,imgIn.getBand(band)->p(cI,rI));
	
	/* Set new output index */
	imgOutIndex[cO+rO*colsOut] = imgInIndex[cI+rI*colsIn][0];
      }
    }
  
    /* Delete variables */
    delete [] rAn;
    delete [] cAn;
    delete [] pOkN;
    delete [] pForN;
    delete [] imgOutN[0];
    delete [] imgOutN;
    delete [] imgInIndex;
    delete [] imgOutIndex;
    
    /* Ouput synthesized texture */
    imgOut.save(opt.getOutputFile());
    
  }
  return 0;
}

/**************************************************************************/

BasicImage::byte pUpdateFn(int cO, int rO, GreyImage &imgProb, int inc) {
  // pUpdateFn MUST return a value x such that 0 < x <= pMax-imgProb.p(cO,rO)
  // and whereby x will equal pMax-imgProb.p(cO,rO) at some point in time.
  BasicImage::byte p = imgProb.p(cO,rO);
  if (inc < pMax-p) return inc;
  return pMax-p;
}

/**************************************************************************/

vector<int> *SimilarPixels(ColourImage &img, GreyImage &imgMask,
			   bool mask, float err) {
  // Return an image of vectors, where each vector is list of indexes to
  // similar pixels for the pixel of concern. Similarity is determine by
  // "err", where two pixels are deemed similar if the L2 (Euclidean) 
  // distance < err. Note even if err = 0, there will always be at least 
  // one pixel in the list, as each pixel is always similar to itself.
  //
  // As the index is a single int,
  // row = index/cols;
  // col = index%cols;
  //
  int rows = img.getHeight();
  int cols = img.getWidth();
  
  /* Setup output index image */
  vector<int> *imgIndex = new vector<int> [cols*rows];
  
  // Instead of collecting a set of indexes for each pixel,
  // the first index will indicate which set of indexes the pixel
  // should refer to. This will stop the repeatitions of sets.
  // Also speed can be improved, as once a pixel has been redirected
  // to a set, it does not need to be looked at again.
  // A pixel can be only redirected to a set, if the pixel value equals
  // the value of the pixel who owns the set, ie when similarity = 0.
  err *= err;
  BasicImage::byte imgBands[3];
  /* Initialize each pixel index */
  for (int r = 0; r < rows; ++r) {
  for (int c = 0; c < cols; ++c) {
    if (!mask || imgMask.p(c,r)) {
      int index=c+r*cols;
      imgIndex[index].push_back(index);
    }
  }
  }

  /* For each pixel */
  for (int r = 0; r < rows; ++r) {
  for (int c = 0; c < cols; ++c) {
    if (!mask || imgMask.p(c,r)) {
      int index=c+r*cols;
      
      /* Check that index has not allready been redirected */
      if (imgIndex[index][0] != index) continue;
      
      /* Get pixel values */
      for (int band = 0; band < 3; ++band)
	imgBands[band] = img.getBand(band)->p(c,r);
    
      /* Search forward */
      int cs = c+1;
      for (int ri = r; ri < rows; ++ri) {
      for (int ci = cs; ci < cols; ++ci, cs = 0) {
	if (!mask || imgMask.p(ci,ri)) {
	  
	  /* Calculate similarity */
	  float similarity = 0.0;
	  for (int band = 0; band < 3; ++band) {
	    float d = imgBands[band]-img.getBand(band)->p(ci,ri);
	    similarity += d*d;
	    if (similarity > err) break;
	  }

	  /* Store new index */
	  if (similarity <= err) {
	    int indexi=ci+ri*cols;
	    imgIndex[index].push_back(indexi);
	    if (similarity == 0.0) imgIndex[indexi].clear();
	    imgIndex[indexi].push_back(index);
	  }
	}
      }
      }
    }
  }
  }

  return imgIndex;
}

/**************************************************************************/

void IndexArray(list<int> &index, int dx, int dy, int cols, int rows,
		GreyImage &imgMask, bool mask, vector<int> &imgIndex) {
  // Update index array with new indexes, making sure that there are no
  // repeatitions and they do not point to points outside the image.
  // The indexes are listed in canological order to allow quick correspondence.

  /* First index */
  /* If imgIndex is neighbourhood pixel then the center pixel is */
  vector<int>::const_iterator im = imgIndex.begin();
  if (im == imgIndex.end()) return;
  int r = ((*im)/cols)-dy;
  int c = ((*im)%cols)-dx;

  /* Check imgIndex with offset is still within the image */
  list<int>::iterator i;
  if ((r >= 0) && (c >= 0) && (c < cols) && (r < rows) &&
      (!mask || imgMask.p(c,r))) {
    
    /* Add imgIndex if it does not already exist */
    int imgIndexi = c+r*cols;
    for (i = index.begin(); i != index.end(); ++i) {
      if (*i == imgIndexi) break;
      if (*i > imgIndexi) {
	index.insert(i,imgIndexi);
	break;
      }
    }
    if (i == index.end()) index.push_back(imgIndexi);
  }

  /* The rest of the indexes are in canological order */
  i = index.begin();
  for (++im; im != imgIndex.end(); ++im) {
    /* If imgIndex is neighbourhood pixel then the center pixel is */
    r = ((*im)/cols)-dy;
    c = ((*im)%cols)-dx;

    /* Check imgIndex with offset is still within the image */
    if ((r < 0) || (c < 0) || (c >= cols) || (r >= rows) ||
	(mask && !imgMask.p(c,r))) continue;
    
    /* Add imgIndex if it does not already exist */
    int imgIndexi = c+r*cols;
    for (; i != index.end(); ++i) {
      if (*i == imgIndexi) break;
      if (*i > imgIndexi) {
	index.insert(i,imgIndexi);
	break;
      }
    }
    if (i == index.end()) index.push_back(imgIndexi);
  }
}

/**************************************************************************/

bool getImgPos(int &c, int &r, list<int> &index, int cols, int rows,
	       GreyImage &imgMask, bool mask) {
  static int i = 0;
  static list<int>::iterator id;
  
  if (index.empty()) {
    int size = cols*rows;
    if (i == size) {
      i = 0;
      return false;
    }
    c = i%cols;
    r = i/cols;
    ++i;
    while (mask && !imgMask.p(c,r)) {
      if (i == size) {
	i = 0;
	return false;
      }
      c = i%cols;
      r = i/cols;
      ++i;
    }
    return true;

  } else {
    if (i == 0) id = index.begin();
    else if (id == index.end()) {
      i = 0;
      return false;
    }
    c = (*id)%cols;
    r = (*id)/cols;
    ++i;
    ++id;
    return true;
  }
}

/**************************************************************************/
 
