#include "ColourImage.h"

#include <iostream>
#include <algorithm>
#include <Magick++.h>

using namespace std;

ColourImage::ColourImage(int w, int h, int lev) :
  BasicImage(w, h, lev) {
  build();
}

ColourImage::ColourImage(const ColourImage& from) :
  BasicImage(from) {
  for (int i=0; i<3; i++) bands[i] = new GreyImage(*from.bands[i]);
}

ColourImage::ColourImage(const string& fname, int lev) {
  for (int i=0; i<3; i++) bands[i] = 0;
  load(fname,lev);
}

ColourImage::~ColourImage() {
  destroy();
}

BasicImage *ColourImage::copy() const {
  return new ColourImage(*this);
}

ColourImage const &ColourImage::operator= (ColourImage const &from) {
  if (this != &from) {
    destroy();
    copy(from);
  }
  return (*this);
}

void ColourImage::resize(int w, int h) {
  destroy();
  width = w;
  height = h;
  size = width*height;
  build();
}

void ColourImage::resize(int w, int h, int levels) {
  quantLevels = levels;
  resize(w,h);
}

bool ColourImage::load(const string& fname) {
  destroy();
  Magick::Image magImg;

  if (!loadMagick(fname, magImg)) return false;
  
  Magick::Blob tmp;
  magImg.write(&tmp, "RGB", 8);

  // create 3 bands
  for (int i=0; i<3; i++) delete bands[i];
  for (int i=0; i<3; i++) {
    bands[i] = new GreyImage(width, height, quantLevels);
  }

  separate(tmp);

  for (int i=0; i<3; i++) {
    bands[i]->quantUniform();
    bands[i]->upsideDown();
  }

  return true;
}

bool ColourImage::load(const string& fname, int lev) {
  quantLevels = lev;
  return load(fname);
}

bool ColourImage::save(const string& fname) const {
  Magick::Blob bl;
  bl.updateNoCopy(new byte[3*size], 3*size);

  for (int i=0; i<3; i++) {
    GreyImage tmp(*bands[i]); // work with copy
    tmp.upsideDown();
    tmp.unquantUniform();

    combine(tmp, i, bl);
  }

  // Magick::Image magImg(bl, Magick::Geometry(width,height), 8, "RGBA");
  Magick::Image magImg;
  magImg.fileName("RGB:dummy"); // ### workaround: ImageMagick format
                                 //       detection broken if filename absent
  magImg.read(bl, Magick::Geometry(width,height), 8);

  return saveMagick(fname, magImg);
}

void ColourImage::writeImageType(ostream& out) const {
  BasicImage::writeImageType(out);
  out << "image_mode " << "colour" << endl;
}

void ColourImage::shuffle() {
  // ### hmm, how to do best?
  cerr << "ColourImage::shuffle: sorry, not yet supported!" << endl;
}

void ColourImage::clear() {
  for (int i=0; i<3; i++) bands[i]->clear();
}

void ColourImage::copy(ColourImage const &from) {
  width = from.width;
  height = from.height;
  size = from.size;
  quantLevels = from.quantLevels;
  for (int i=0; i<3; i++) bands[i] = new GreyImage(*from.bands[i]);
}

void ColourImage::destroy() {
  for (int i=0; i<3; i++) {
    delete bands[i];
    bands[i] = 0;
  }
}

void ColourImage::build() {
  for (int i=0; i<3; i++) {
    bands[i] = new GreyImage(width, height, quantLevels);
  }
}

void ColourImage::separate(const Magick::Blob& bl) {
  byte *p = (byte *)bl.data();
  for (int y = 0; y < height; ++y) {
  for (int x = 0; x < width; ++x) {
    for (int j=0; j<3; ++j) {
      bands[j]->setp(x, y, *p++);
    }
  }
  }
}

void ColourImage::combine(const GreyImage& band, int n,
			  Magick::Blob& target) const {
  byte *p = (byte *)target.data() + n;
  for (int y = 0; y < height; ++y) {
  for (int x = 0; x < width; ++x) {
    *p = band.p(x,y);
    p += 3;
  }
  }
}

ColourImage const &ColourImage::reduce(ColourImage const &from) {
  width = (from.width+1)/2;
  height = (from.height+1)/2;
  size = width*height;
  quantLevels = from.quantLevels;
  for (int i=0; i<3; i++) bands[i]->reduce(*(from.bands[i]));
  return (*this);
}

ColourImage const &ColourImage::sample(ColourImage const &from) {
  width = (from.width+1)/2;
  height = (from.height+1)/2;
  size = width*height;
  quantLevels = from.quantLevels;
  for (int i=0; i<3; i++) bands[i]->sample(*(from.bands[i]));
  return (*this);
}

